/* File: demag2.cc */
/*
 * (New) Constant magnetization demag routines, based on formulae
 * presented in "A Generalization of the Demagnetizing Tensor for
 * Nonuniform Magnetization," by Andrew J. Newell, Wyn Williams, and
 * David J. Dunlop, Journal of Geophysical Research, vol 98, p 9551-9555,
 * June 1993.  This formulae clearly satisfy necessary symmetry and
 * scaling properties, which is not true of the formulae presented in
 * "Magnetostatic Energy Calculations in Two- and Three-Dimensional
 * Arrays of Ferromagnetic Prisms," M. Maicas, E. Lopez, M. CC. Sanchez,
 * C. Aroca and P. Sanchez, IEEE Trans Mag, vol 34, May 1998, p601-607.
 * (Side note: The units in the latter paper are apparently cgs.)  It
 * appears likely that there is an error in the latter paper (attempts
 * to implement the formulae presented there did not produce the proper
 * symmetries), as well as in the older paper, "Magnetostatic
 * Interaction Fields for a Three-Dimensional Array of Ferromagnetic
 * Cubes," Manfred E. Schabes and Amikam Aharoni, IEEE Trans Mag, vol
 * 23, November 1987, p3882-3888.  (Note: The Newell paper deals with
 * uniformly sized rectangular prisms, the Maicas paper allows
 * non-uniformly sized rectangular prisms, and the Schabes paper only
 * considers cubes.)
 *
 *   The kernel here is based on an analytically derived energy, and the
 * effective (discrete) demag field is calculated from the (discrete)
 * energy.
 *
 */

#include <math.h>
#include "nb.h"

#include "demag.h"
#include "fft.h"

#ifdef USE_MPI
#include "mmsmpi.h"
#endif /* USE_MPI */

/* End includes */

DF_Init     ConstMagInit;
DF_Destroy  ConstMagDestroy;
DF_Calc     ConstMagField;

static double Aspect(1.0);  // film thickness/cellsize
static int Nx(1),Ny(1);     // Input array dimensions, space domain
static int NFx(1),NFy(1);   // Array dimensions, in freq domain

// Kernel arrays, in frequency domain.  See MJD NOTES, 13-Dec-1998, p189
// for details.  The k,l superscripts correspond to the coordinate
// directions, with 0=>x, 1=>y, 2=>z.
static MyComplex **A00 = NULL;
static MyComplex **A01 = NULL;
static MyComplex **A11 = NULL;
static MyComplex **A22 = NULL;		 /// Other terms are zero.
// Magnetization arrays
static double **spacework = NULL;   // Space domain workspace
static MyComplex **freqwork = NULL; // Freqency domain workspace
static MyComplex **M0 = NULL;
static MyComplex **M1 = NULL;
static MyComplex **M2 = NULL;    // Frequeny domain
                                /// magnetizations, M0 is x-component,
                                /// M1 is y-component, M2 is z-component.

#ifndef USE_MPI
static FFTReal2D *fft = NULL;
// Class to handle forward and reverse FFT's.
#else
static FFTReal2D_mpi *fft = NULL;
// Class to handle forward and reverse FFT's, MPI-style.
// NOTE: We make this a dynamically allocated beastie so it can
//       be dynamically de-allocated _before_ MPI_Finalize() is
//       called.
#endif

////////////////////////////////////////////////////////////////////////////
// Routines to calculate kernel coefficients
// See Newell et al. for details. The code below follows the
// naming conventions in that paper.
static double
f(double x,double y,double z)
{ // There is mucking around here to handle case where imports
  // are near zero.  In particular, asinh(t) is written as
  // log(t+sqrt(1+t)) because the latter appears easier to
  // handle if t=y/x (for example) as x -> 0.

 // This function is even; the fabs()'s just simplify special case handling.
  x=fabs(x); double xsq=x*x;
  y=fabs(y); double ysq=y*y;
  z=fabs(z); double zsq=z*z; 

  double R=xsq+ysq+zsq;
  if(R<=0.0) return 0.0;
  else       R=sqrt(R);

  // f(x,y,z)
  double result=0.;
  if(z>0.) { // For 2D grids, half the calls from F1 have z==0.
    double temp1,temp2,temp3;
    result = (1./6.)*(2*xsq-ysq-zsq)*R;
    if((temp1=x*y*z)>0.)
      result += -temp1*atan2(y*z,x*R);
    if(y>0. && (temp2=xsq+zsq)>0.)
      result += 0.5*y*(zsq-xsq)*log((y+R)/sqrt(temp2));
    if((temp3=xsq+ysq)>0.)
      result += 0.5*z*(ysq-xsq)*log((z+R)/sqrt(temp3));
  } else {
    // z==0
    result = (1./6.)*(2*xsq-ysq)*R;
    if(y>0. && x>0.)
      result += -0.5*y*xsq*log((y+R)/x);
  }

  return result;
}

static double
SelfDemagDx(double x,double y,double z)
{ // Note: Assumes x, y, and z are all >0.
  // Note 2: egcs-2.91.57 on Linux/x86 with -O1 mangles this
  //  function (produces NaN's) unless we manually group terms.

  double xsq=x*x,ysq=y*y,zsq=z*z;
  double diag=sqrt(xsq+ysq+zsq);

  double temp1=
    - (2*xsq*x-ysq*y-zsq*z)
    + (2*xsq-ysq)*sqrt(xsq+ysq)
    + (2*xsq-zsq)*sqrt(xsq+zsq)
    - (ysq+zsq)*sqrt(ysq+zsq)
    - (2*xsq-ysq-zsq)*diag;
  temp1*=(4./3.);

  double temp2=
    + 2*x*y*z*atan(y*z/(x*diag))
    + (z+y)*xsq*log(x);
  temp2*=4;

  double temp3=
    + 2*(ysq*(log((sqrt(ysq+zsq)+z)/y))
	 - xsq*log(sqrt(xsq+zsq)+z))
    + (xsq-ysq)*(2*log(diag+z) - log(xsq+ysq));
  temp3*=2*z;

  double temp4=
    + 2*(zsq*(log((sqrt(ysq+zsq)+y)/z))
	 - xsq*log(sqrt(xsq+ysq)+y))
    + (xsq-zsq)*(2*log(diag+y) - log(xsq+zsq));
  temp4*=2*y;

  double Dx = temp1 + temp2 + temp3 + temp4;

  return Dx;
}

static double
GetSDA00(double x,double y,double z,double dx,double dy,double dz)
{ // This is Nxx/(4*PI*tau) in Newell's paper
  double result=0.;
  if(x==0. && y==0. && z==0.) {
    // Self demag term.  The base routine can handle x==y==z==0,
    // but this should be more accurate.
    result = SelfDemagDx(dx,dy,dz);
  } else {
    // Simplified (collapsed) formula based on Newell's paper.
    // This saves about half the calls to f().  There is still
    // quite a bit of redundancy from one cell site to the next,
    // but as this is an initialization-only issure speed shouldn't
    // be too critical.
    double result1,result2,result4,result8;
    result1 =
      -1*(f(x+dx,y+dy,z+dz)+f(x+dx,y-dy,z+dz)+f(x+dx,y-dy,z-dz)+
	  f(x+dx,y+dy,z-dz)+f(x-dx,y+dy,z-dz)+f(x-dx,y+dy,z+dz)+
	  f(x-dx,y-dy,z+dz)+f(x-dx,y-dy,z-dz));
    result2 =
      2*(f(x,y-dy,z-dz)+f(x,y-dy,z+dz)+f(x,y+dy,z+dz)+f(x,y+dy,z-dz)+
	 f(x+dx,y+dy,z)+f(x+dx,y,z+dz)+f(x+dx,y,z-dz)+f(x+dx,y-dy,z)+
	 f(x-dx,y-dy,z)+f(x-dx,y,z+dz)+f(x-dx,y,z-dz)+f(x-dx,y+dy,z));
    result4 =
      -4*(f(x,y-dy,z)+f(x,y+dy,z)+f(x,y,z-dz)+
	  f(x,y,z+dz)+f(x+dx,y,z)+f(x-dx,y,z));
    result8 =
      8*f(x,y,z);
    result=result1+result2+result4+result8;
  }
  return result;
  /// Multiply result by 1./(4*PI*dx*dy*dz) to get effective "demag"
  /// factor Nxx from Newell's paper.
}

static inline double
GetSDA11(double x,double y,double z,double l,double h,double e)
{ return GetSDA00(y,z,x,h,e,l); }

static inline double
GetSDA22(double x,double y,double z,double l,double h,double e)
{ return GetSDA00(z,x,y,e,l,h); }

static double
g(double x,double y,double z)
{ // There is mucking around here to handle case where imports
  // are near zero.  In particular, asinh(t) is written as
  // log(t+sqrt(1+t)) because the latter appears easier to
  // handle if t=y/x (for example) as x -> 0.

  double result_sign=1.0;
  if(x<0.0) result_sign *= -1.0;  if(y<0.0) result_sign *= -1.0;
  x=fabs(x); y=fabs(y); z=fabs(z);  // This function is even in z and
  /// odd in x and y.  The fabs()'s simplify special case handling.

  double xsq=x*x,ysq=y*y,zsq=z*z;
  double R=xsq+ysq+zsq;
  if(R<=0.0) return 0.0;
  else       R=sqrt(R);

  // g(x,y,z)
  double result = -2*x*y*R;;
  if(z>0.) { // For 2D grids, 1/3 of the calls from GetSDA01 have z==0.
    double temp1,temp2,temp3;
    result += 
      -z*(zsq*atan2(x*y,z*R)
	  +3.*(ysq*atan2(x*z,y*R)+xsq*atan2(y*z,x*R)));

    if((temp1=xsq+ysq)>0.)
      result += 6*x*y*z*log((z+R)/sqrt(temp1));

    if((temp2=ysq+zsq)>0.)
      result += y*(3*zsq-ysq)*log((x+R)/sqrt(temp2));

    if((temp3=xsq+zsq)>0.)
      result += x*(3*zsq-xsq)*log((y+R)/sqrt(temp3));

  } else {
    // z==0.
    if(y>0.) result += -y*ysq*log((x+R)/y);
    if(x>0.) result += -x*xsq*log((y+R)/x);
  }

  return (1./6.)*result_sign*result;

}

static double
GetSDA01(double x,double y,double z,double l,double h,double e)
{ // This is Nxy/(4*PI*tau) in Newell's paper.

  // Simplified (collapsed) formula based on Newell's paper.
  // This saves about half the calls to g().  There is still
  // quite a bit of redundancy from one cell site to the next,
  // but as this is an initialization-only issure speed shouldn't
  // be too critical.

  double result1 =
    -1*(g(x-l,y-h,z-e)+g(x-l,y-h,z+e)+g(x+l,y-h,z+e)+g(x+l,y-h,z-e)+
	g(x+l,y+h,z-e)+g(x+l,y+h,z+e)+g(x-l,y+h,z+e)+g(x-l,y+h,z-e));

  double result2 =
    2*(g(x,y+h,z-e)+g(x,y+h,z+e)+g(x,y-h,z+e)+g(x,y-h,z-e)+
       g(x-l,y-h,z)+g(x-l,y+h,z)+g(x-l,y,z-e)+g(x-l,y,z+e)+
       g(x+l,y,z+e)+g(x+l,y,z-e)+g(x+l,y-h,z)+g(x+l,y+h,z));

  double result4 =
    -4*(g(x-l,y,z)+g(x+l,y,z)+g(x,y,z+e)+g(x,y,z-e)+
	g(x,y-h,z)+g(x,y+h,z));

  double result8 =
    8*g(x,y,z);

  return result1+result2+result4+result8;
  /// Multiply result by 1./(4*PI*l*h*e) to get effective "demag"
  /// factor Nxy from Newell's paper.
}


////////////////////////////////////////////////////////////////////////////

static int NextPowerOfTwo(int n)
{ // Helper function.  Returns first power of two >= n
  int m=1;
  while(m<n) m*=2;
  return m;
}

// A couple helper macros
#define ALLOC2D(ARRAY,XDIM,YDIM,TYPE,I)                     \
           ARRAY=new TYPE*[XDIM];                           \
           ARRAY[0]=new TYPE[XDIM*YDIM];                    \
           for(I=1;I<XDIM;I++) ARRAY[I]=ARRAY[I-1]+YDIM;

#define DEALLOC2D(ARRAY) \
      if(ARRAY!=NULL) { delete[] ARRAY[0]; delete[] ARRAY; ARRAY=NULL; }


static void DeallocateArrays()
{ // Release memory used by file static arrays
  DEALLOC2D(spacework);
  DEALLOC2D(freqwork);
  DEALLOC2D(M0);   DEALLOC2D(M1);   DEALLOC2D(M2);
  DEALLOC2D(A00);  DEALLOC2D(A01);  DEALLOC2D(A11);  DEALLOC2D(A22);
}


static void AllocateArrays(int xdim,int ydim)
{ // Allocate memory for file static arrays, and sets Nx, Ny, NFx & NFy.
  DeallocateArrays();
  Nx=xdim;  Ny=ydim;

  // To get the correct "aperiodic" convolution from the FFT's,
  // it appears that the kernel needs to be expanded about zero,
  // and the input zero padded, to N + (N-1).  See the notes in
  // fft.h concerning the implicit zero padding of FFT's of real
  // functions.
  NFy=NextPowerOfTwo(Ny+Ny-1);
  NFx=(NextPowerOfTwo(Nx+Nx-1))/2+1; // For real inputs, the resulting
  /// complex transform is symmetric, so only the top half is computed.

  int i;
  ALLOC2D(spacework,Nx,Ny,double,i);

  ALLOC2D(freqwork,NFx,NFy,MyComplex,i);
  ALLOC2D(M0,NFx,NFy,MyComplex,i);
  ALLOC2D(M1,NFx,NFy,MyComplex,i);
  ALLOC2D(M2,NFx,NFy,MyComplex,i);

  ALLOC2D(A00,NFx,NFy,MyComplex,i);
  ALLOC2D(A01,NFx,NFy,MyComplex,i);
  ALLOC2D(A11,NFx,NFy,MyComplex,i);
  ALLOC2D(A22,NFx,NFy,MyComplex,i);
}

void
ConstMagInit(const int xdim,const int ydim,
	     int argc, char** argv)
{ // Allocates memory, calculates kernels, and sets global values
  // argv[0] should be "ConstMag", and argv[1] should be the ratio
  // film thickness/cellsize (Aspect).
  if(xdim<1 || ydim<1 || argc!=2 ||
     Nb_StrCaseCmp(argv[0],"ConstMag")!=0) {
    PlainError(1,
	    "FATAL ERROR: Invalid argument "
	    "(->%s<- with argc=%d, dim=(%d,%d))"
	    " to ConstMagInit()\n",argv[0],argc,xdim,ydim);
  }
#ifndef USE_MPI
  fft=new FFTReal2D;
#else
  fft=new FFTReal2D_mpi;
#endif
  Aspect=Nb_Atof(argv[1]);
  AllocateArrays(xdim,ydim);

  // Initialize kernel
  int i,j;
#ifdef DEBUG_MPI
#define R1 5
#define C1 6
#define R2 5
#define C2 8
  double *test[R1];      double testarr[R1*C1];
  test[0]=testarr; for(i=1;i<R1;i++) test[i]=test[i-1]+C1;
  MyComplex *ctest[R2];  MyComplex ctestarr[R2*C2];
  ctest[0]=ctestarr; for(i=1;i<R2;i++) ctest[i]=ctest[i-1]+C2;
  for(i=0;i<R1;i++) for(j=0;j<C1;j++) test[i][j]=1.0+i+j;
  for(i=0;i<R2;i++) for(j=0;j<C2;j++) ctest[i][j]=MyComplex(-77.7777,-55.5555);
  printf("Initial array---\n");
  for(i=0;i<R1;i++) {
    for(j=0;j<C1;j++) printf("%10.5f ",test[i][j]);
    printf("\n");
  }
  fft->Forward(R1,C1,test,R2,C2,ctest);
  printf("\nTransformed array---\n");
  for(i=0;i<R2;i++) {
    for(j=0;j<C2;j++)
      printf("(%6.2f,%6.2f) ",ctest[i][j].real(),ctest[i][j].imag());
    printf("\n");
  }
  for(i=0;i<R1;i++) for(j=0;j<C1;j++) test[i][j]=-42.424242;
  fft->Inverse(R2,C2,ctest,R1,C1,test);
  printf("F+I array---\n");
  for(i=0;i<R1;i++) {
    for(j=0;j<C1;j++) printf("%10.5f ",test[i][j]);
    printf("\n");
  }
#endif /* DEBUG_MPI */

  int Kx=2*(NFx-1);  int Ky=NFy;  // Space domain kernel dimensions
  double **spacekernel;
  ALLOC2D(spacekernel,Kx,Ky,double,i);
  for(i=0;i<Kx;i++) for(j=0;j<Ky;j++) spacekernel[i][j]=0.;  // Zero fill

  double l,h,e,z,scale;
  l=h=1.0; e=Aspect; z=0.0;  // Scale size so distance between adjacent
                            /// cells is 1.
  scale= -1./(4*PI*l*h*e);
  /// According (16) in Newell's paper, the demag field is given by
  ///                        H = -N*M
  /// where N is the "demagnetizing tensor," with components Nxx, Nxy,
  /// etc.  With the '-1' in 'scale' we store '-N' instead of 'N',
  /// so we don't have to multiply the output from the FFT + iFFT
  /// by -1 in ConstMagField() below.

  for(i=0;i<Nx;i++) for(j=0;j<Ny;j++) {
    spacekernel[i][j]=scale*GetSDA00(double(i),double(j),z,l,h,e);
    if(i>0) spacekernel[Kx-i][j]=spacekernel[i][j];
    if(j>0) spacekernel[i][Ky-j]=spacekernel[i][j];
    if(i>0 && j>0) spacekernel[Kx-i][Ky-j]=spacekernel[i][j];
  }
  fft->Forward(Kx,Ky,(const double* const*)spacekernel,NFx,NFy,A00);

  for(i=0;i<Nx;i++) for(j=0;j<Ny;j++) {
    spacekernel[i][j]=scale*GetSDA11(double(i),double(j),z,l,h,e);
    if(i>0) spacekernel[Kx-i][j]=spacekernel[i][j];
    if(j>0) spacekernel[i][Ky-j]=spacekernel[i][j];
    if(i>0 && j>0) spacekernel[Kx-i][Ky-j]=spacekernel[i][j];
  }
  fft->Forward(Kx,Ky,(const double* const*)spacekernel,NFx,NFy,A11);

  for(i=0;i<Nx;i++) for(j=0;j<Ny;j++) {
    spacekernel[i][j]=scale*GetSDA22(double(i),double(j),z,l,h,e);
    if(i>0) spacekernel[Kx-i][j]=spacekernel[i][j];
    if(j>0) spacekernel[i][Ky-j]=spacekernel[i][j];
    if(i>0 && j>0) spacekernel[Kx-i][Ky-j]=spacekernel[i][j];
  }
  fft->Forward(Kx,Ky,(const double* const*)spacekernel,NFx,NFy,A22);

  for(i=0;i<Nx;i++) for(j=0;j<Ny;j++) {
    spacekernel[i][j]=scale*GetSDA01(double(i),double(j),z,l,h,e);
    if(i>0) spacekernel[Kx-i][j]= -1.0*spacekernel[i][j];
    if(j>0) spacekernel[i][Ky-j]= -1.0*spacekernel[i][j];
    if(i>0 && j>0) spacekernel[Kx-i][Ky-j]=spacekernel[i][j];
    // A01 is odd in x and y.
  }
  fft->Forward(Kx,Ky,(const double* const*)spacekernel,NFx,NFy,A01);
  DEALLOC2D(spacekernel);
}

void
ConstMagDestroy(void)
{
  DeallocateArrays();
  delete fft;
  fft=NULL;
}

void
ConstMagField(DF_Export_m* writem,DF_Import_h* readh,
	      void* m,void* h)
{ // Import functions writem and readh must be provided by caller,
  // and provide access into the hidden magnetization and field
  // structures.
  int i,j;

  // Take FFT of magnetization components.
  writem(spacework,m,Nx,Ny,0);
  fft->Forward(Nx,Ny,(const double* const*)spacework,NFx,NFy,M0);
  writem(spacework,m,Nx,Ny,1);
  fft->Forward(Nx,Ny,(const double* const*)spacework,NFx,NFy,M1);
  writem(spacework,m,Nx,Ny,2);
  fft->Forward(Nx,Ny,(const double* const*)spacework,NFx,NFy,M2);

  // Calculate H0 component
  for(i=0;i<NFx;i++) for(j=0;j<NFy;j++) {
    freqwork[i][j]=A00[i][j]*M0[i][j]+A01[i][j]*M1[i][j];
  }
  fft->Inverse(NFx,NFy,(const MyComplex* const*)freqwork,Nx,Ny,spacework);
  readh(spacework,h,Nx,Ny,0);

  // Calculate H1 component
  for(i=0;i<NFx;i++) for(j=0;j<NFy;j++) {
    freqwork[i][j]=A01[i][j]*M0[i][j]+A11[i][j]*M1[i][j];
  }
  fft->Inverse(NFx,NFy,(const MyComplex* const*)freqwork,Nx,Ny,spacework);
  readh(spacework,h,Nx,Ny,1);

  // Calculate H2 component
  for(i=0;i<NFx;i++) for(j=0;j<NFy;j++) {
    freqwork[i][j]=A22[i][j]*M2[i][j];
  }
  fft->Inverse(NFx,NFy,(const MyComplex* const*)freqwork,Nx,Ny,spacework);
  readh(spacework,h,Nx,Ny,2);
}
